/* 20220121
 *  Corrected Phase Marking
 *  
 *  
 *  
 *  20220114
 * Implemented Mute button
 * Correncted EEPROM for int16_t
 * 
 * 
 * 
 * 20220112: Implemented Remote Command
 *            Implemented EEPROM SAVE
 *            Introduced DEBUG options
 *            Cleanup code
 *            
 * 
 *
 * 
 * 20210601: Implemented left and right chammel
 *            Inizialization of muses first
 *            Display off in Standby
 *            GAIN definition
 *            IR pin change now 3 (interrupt)
 *            Corrected volume index bug
 *            Disable knob on standby
 *            Pre on with single click
 *            Serial debug
 *            Refined menu with cursors in offset
 */
/*
 Parameter Description
 0 Main Volume
 1 Right Offset
 2 Left Offset
 3 Right Plus Offset
 4 Right Minus Offset
 5 Left Plus Offset
 6 Left Minus Offset 
 
 10 Source1 Right Minus Offset
 11 Source1 Right Plus Offset
 12 Source1 Left Minus Offset
 13 Source1 Left Plus Offset 
 *
 20 Source1 Right Minus Offset
 21 Source1 Right Plus Offset
 22 Source1 Left Minus Offset
 23 Source1 Left Plus Offset 
 * 
 * ...
 
 */

// The array holds the parameters for each input
// (The current input selection is recorded after the array)
//int16_t settings[MAX_PARAMETER+1];  // Array to hold parameter values




//Libraries                   

//Encoder
#include <ClickEncoder.h>
#include <TimerOne.h>
//Muses
#include <Muses72323.h>
//LCD
#include <LiquidCrystal_I2C.h>
//EEPROM
#include <EEPROM.h>
//Remote
#include <Arduino.h>
//#include "PinDefinitionsAndMore.h"
#include <IRremote.hpp>

String build = "    20220121_001    "; 
String date =  "  21 January 2022   ";

//#define DEBUG 1

#define B 0xFF  // The character for a completely filled box
#define A 0x20  // The character for blank

//PIN Assignment
//input selection OUT
#define INPUT_1 4 // Yellow
#define INPUT_2 5 //Orange
#define INPUT_3 6 // Red
#define INPUT_4 7 // Brown 
//power control OUT
#define PRE_ON 8 // Green Blue Pin 1 on PSU
#define AMP_ON 9 //Orange-blue Pin2(Left) and Pin 2(Right) on PSU
//Cabling color from Arduino to Board
//10 Yellow-Yellow-Yellow   (SS-SDL)
//11 Orange-red-Blue        (MOSI-SDA)
//12 Empty                   (MISO)
//13 Green-Green-Green      (SCK-SCL)
#define IR_RECEIVE_PIN 3 // Blue
#define MAIN_MENU 0
#define SOURCE1_MENU 1
#define SOURCE2_MENU 2
#define SOURCE3_MENU 3
#define SOURCE4_MENU 4

#define STANDBY 0
#define PREON   1
#define AMPON   2

#define MAX_MENU        4
#define MAX_INPUT       3
#define MAX_PARAMETER   46
#define MAX_ATTENUATION -472
#define MAX_STATUS       2
#define STEP             0.25 
#define GAIN              10
#define REMOTE_DELAY    200
#define LONG_COUNT      10

// Set 1 Denon TU Remote
#define IR_UP       0xC5
#define IR_DOWN     0x25
#define IR_LEFT     0xFE
#define IR_RIGHT    0xBB
#define IR_ENTER    0x7
#define IR_1        0x83
#define IR_2        0x43
#define IR_3        0xC3
#define IR_4        0x23
#define IR_5        0xA3
#define IR_6        0x63
#define IR_7        0xE3
#define IR_8        0x13
#define IR_9        0x93
#define IR_0        0x53
#define MUTE        0xF

// The address wired into the muses chip (usually 0).
static const byte MUSES_ADDRESSR = 0;
static const byte MUSES_ADDRESSL = 1;

static Muses72320 MusesR(MUSES_ADDRESSR);
static Muses72320 MusesL(MUSES_ADDRESSL);

static Muses72320::volume_t CurrentVolume = -80;

ClickEncoder *main_knob;

int16_t parameters[MAX_PARAMETER+1] ;
int16_t parameters_max_limit[MAX_PARAMETER+1] ;
int16_t parameters_min_limit[MAX_PARAMETER+1] ;

int16_t last[MAX_PARAMETER+1] ;

LiquidCrystal_I2C lcd(0x27,20,4);

int inputs[MAX_INPUT+1]={INPUT_1,INPUT_2,INPUT_3,INPUT_4,};
String input_name[MAX_INPUT+1] = {"Denon  ","9018 D1","9018SEN","AD1865 "};
//int input_volume[MAX_INPUT+1]={0,0,0,30};
//word settings[INPUTS_NUMBER];  // Array to hold parameter values

bool mute = false ;

int input_index= 0 ;
int parameter_index=0;
int status_index=STANDBY; //0 STANDBY 1 PRE-ON 2 AMP ON
int menu_index=0;
int count=0;
/*
unsigned long max_time = 0;
unsigned long inactivity_time = 0 ;
unsigned long activity_time = 0 ;
unsigned long warmup_time = 0 ;
unsigned long pre_time = 0
*/
void timerIsr() {
  main_knob->service();
}

void writeIntIntoEEPROM(int address, int16_t number)
{ 
  byte byte1 = number >> 8;
  byte byte2 = number & 0xFF;

  EEPROM.write(address, byte1);
  EEPROM.write(address + 1, byte2);
}

int readIntFromEEPROM(int address)
{
  byte byte1 = EEPROM.read(address);
  byte byte2 = EEPROM.read(address + 1);

  return (byte1 << 8) + byte2;
}

void initialize_parameters()
{
for (int i=0; i <MAX_PARAMETER;i++)
    {
         if (parameters[i] > 0 ) parameters[i]=0;
         parameters_min_limit[i]= MAX_ATTENUATION;
         parameters_max_limit[i]= 0;
         last[1]=-1;
         if (parameters[0] > 0) parameters[0]=CurrentVolume;  
    } 
}
// Routines for reading and writing from EEPROM

void writeParameters(){
   #ifdef DEBUG
     Serial.println("WRITE");
   #endif
  int eeAddress = 0;
  for(int i=0;i<MAX_PARAMETER;i++) {
    if(readIntFromEEPROM(i)!=abs(parameters[i])) // Check an see if there are any changes
    {   
      #ifdef DEBUG
        Serial.print(i);
        Serial.print(" ");
      #endif
        writeIntIntoEEPROM(eeAddress,abs(parameters[i]));
        #ifdef DEBUG
        Serial.print(parameters[i]);
        Serial.print(" ");
        Serial.println(EEPROM.read(eeAddress));
        #endif
        
    }
    eeAddress += sizeof(int16_t);
  } 
}


void readParameters(){
   #ifdef DEBUG 
    Serial.println("READ");
   #endif
  int eeAddress = 0;
  for (int i=0;i<MAX_PARAMETER;i++) 
  {
      #ifdef DEBUG
          Serial.print(i);
          Serial.print(" ");
          Serial.print( readIntFromEEPROM(eeAddress));
      #endif
      parameters[i] = (readIntFromEEPROM(eeAddress) * -1 );
      
      #ifdef DEBUG
          Serial.print(" ");
          Serial.println(parameters[i]);
      #endif
      
      eeAddress += sizeof(int16_t);
    }
     
}

//Muses Related Functions
void SetupMuses()
{
 // Initialize muses (SPI, pin modes)...
  MusesR.setExternalClock(false); //Must be set !!!
  MusesR.setZeroCrossingOn(true);
  MusesL.setExternalClock(false); //Must be set !!!
  MusesL.setZeroCrossingOn(true);
}

void muted()
{
MusesR.setVolume(MAX_ATTENUATION,MAX_ATTENUATION);
MusesL.setVolume(MAX_ATTENUATION,MAX_ATTENUATION);
mute = true;
}

void Update_Volume()
{
  int16_t RP,RM,LP,LM;
  RP=parameters[0]+parameters[((input_index+1)*10)];
  RM=parameters[0]+parameters[((input_index+1)*10)+1];
  LP=parameters[0]+parameters[((input_index+1)*10)+2];
  LM=parameters[0]+parameters[((input_index+1)*10)+3];
  if ( RP < MAX_ATTENUATION ) RP = MAX_ATTENUATION;
  if ( RM < MAX_ATTENUATION ) RM = MAX_ATTENUATION;
  if ( LP < MAX_ATTENUATION ) LP = MAX_ATTENUATION;
  if ( LM < MAX_ATTENUATION ) LM = MAX_ATTENUATION;
  MusesR.setVolume(RP,RM);
  MusesL.setVolume(LP,LM);
  #ifdef DEBUG
    Serial.print("\n");
    Serial.print(LP);
    Serial.print("\t");
    Serial.print(LM);
    Serial.print("\t");
    Serial.print(RP);
    Serial.print("\t");
    Serial.print(RM);
    Serial.print("\n");
    printpar();
  #endif
  
  mute=false ;
}

//Encoder Related Functions
void main_knobf()
 { 

 int increment = main_knob->getValue();
 parameters[parameter_index] += increment ;
 if (parameter_index == 0) parameters[0] += 3 *  increment ;
  
  if (parameters[parameter_index] != last[parameter_index]) 
  {
    last[parameter_index] = parameters[parameter_index];
    if (parameters[parameter_index] > parameters_max_limit[parameter_index] ) parameters[parameter_index] = parameters_max_limit[parameter_index];
    if (parameters[parameter_index] < parameters_min_limit[parameter_index] ) parameters[parameter_index] = parameters_min_limit[parameter_index];

    writeParameters();
    Update_Volume();
    refresh_display();
  }
 }

void main_button()
{
  ClickEncoder::Button b = main_knob->getButton();
  if (b != ClickEncoder::Open) {
   //Serial.print("Button: ");
    switch (b) {
      case ClickEncoder::Clicked:
      if (status_index == STANDBY) 
      {
        status_index=PREON;
        status_Update();
       }
      else
      {
            if (menu_index>0)
            {
              parameter_index++;
              if (parameter_index>((menu_index*10)+3)) parameter_index=(menu_index*10);
              lcd.clear();
              refresh_display();   
            }else
                  {
                     input_index++;
                      if (input_index>MAX_INPUT) input_index=0; 
                      select_Input(input_index);
                      lcd.clear();
                      refresh_display();
                  }
      }
        break;
        
        case ClickEncoder::DoubleClicked:
        
         menu_index++;
         if (menu_index>MAX_MENU) menu_index=0;
         parameter_index=menu_index*10;
         if (parameter_index>MAX_PARAMETER) parameter_index=0;
         lcd.clear();
         refresh_display();
         break;
         
       case ClickEncoder::Held:
      // case ClickEncoder::Pressed:
      // case ClickEncoder::Released:
          status_index++;
          if (status_index>MAX_STATUS) status_index=0;  
          status_Update();
          refresh_display();
          delay(500);
       break;
  
    }}
  }  
 
void remote()
{
  if (IrReceiver.decode()) {
      delay(REMOTE_DELAY);
      IrReceiver.resume();

        switch (IrReceiver.decodedIRData.command) {
        case (IR_ENTER):
          Serial.println("ENTER");
   
          if (menu_index>0){
            menu_index=0;
            parameter_index=0;
            lcd.clear();
            refresh_display(); 
          }else
          {

          count++;
          if (count == LONG_COUNT)
            {
            status_index++;
            if (status_index>MAX_STATUS) status_index=0;  
            status_Update();
            refresh_display();
            delay(500);
            count = 0;
            }
          }
       break;
       case (IR_1)://1
          if (menu_index>0)
            {
              parameter_index = menu_index*10 ;
              lcd.clear();
              refresh_display();   
            }else
                  {
                     input_index = 0; 
                      select_Input(input_index);
                      lcd.clear();
                      refresh_display();
                  }
       break;

       case (IR_2)://2
          if (menu_index>0)
            {
              parameter_index = menu_index*10 + 1;
              lcd.clear();
              refresh_display();   
            }else
                  {
                     input_index = 1; 
                      select_Input(input_index);
                      lcd.clear();
                      refresh_display();
                  }
       break;

       case (IR_3)://3
           if (menu_index>0)
            {
              parameter_index = menu_index*10 + 2;
              lcd.clear();
              refresh_display();   
            }else
                  {
                     input_index = 2; 
                      select_Input(input_index);
                      lcd.clear();
                      refresh_display();
                  }
       break;

       case (IR_4)://4
           if (menu_index>0)
            {
              parameter_index = menu_index*10 + 3;
              lcd.clear();
              refresh_display();   
            }else
                  {
                     input_index = 3; 
                      select_Input(input_index);
                      lcd.clear();
                      refresh_display();
                  }
       break;

       case (IR_RIGHT): //RIGHT
        #ifdef DEBUG
          Serial.println("RIGHT");
        #endif
         menu_index++;
         if (menu_index>MAX_MENU) menu_index=0;
         parameter_index=menu_index*10;
         if (parameter_index>MAX_PARAMETER) parameter_index=0;
         lcd.clear();
         refresh_display();
         break;
        
        case (IR_LEFT): //LEFT
         #ifdef DEBUG
          Serial.println("LEFT");
         #endif
         menu_index--;
         if (menu_index<0) menu_index=MAX_MENU;
         parameter_index=menu_index*10;
         if (parameter_index>MAX_PARAMETER) parameter_index=0;
         lcd.clear();
         refresh_display();
         break;

         case (IR_UP)://UP
          { 
             parameters[parameter_index] ++;
             if (parameter_index == 0) parameters[0] +=3 ;
              
              if (parameters[parameter_index] != last[parameter_index]) 
              {
                last[parameter_index] = parameters[parameter_index];
                if (parameters[parameter_index] > parameters_max_limit[parameter_index] ) parameters[parameter_index] = parameters_max_limit[parameter_index];
                if (parameters[parameter_index] < parameters_min_limit[parameter_index] ) parameters[parameter_index] = parameters_min_limit[parameter_index];

                writeParameters();
                Update_Volume();
                refresh_display();
              }
          }
          break;

           case (IR_DOWN)://DOWN
          { 
             parameters[parameter_index] --;
             if (parameter_index == 0) parameters[0] -= 3 ;
              
              if (parameters[parameter_index] != last[parameter_index]) 
              {
                last[parameter_index] = parameters[parameter_index];
                if (parameters[parameter_index] > parameters_max_limit[parameter_index] ) parameters[parameter_index] = parameters_max_limit[parameter_index];
                if (parameters[parameter_index] < parameters_min_limit[parameter_index] ) parameters[parameter_index] = parameters_min_limit[parameter_index];

                writeParameters();
                Update_Volume();
                refresh_display();
              }   
            }
          break;

        case(MUTE):
  
           if (mute) 
           {  
            Update_Volume();
            mute = false ;
           }
           else
           {
            muted();
            mute = true;
           }
           refresh_display();
         break;
          
        #ifdef DEBUG
         case(IR_5):

            readParameters();

         break;

         case (IR_6):
              writeParameters();
         break;

        #endif
        }//end switch
      
  
  }//end decode()
  }//end remote()

//Display Related Functions
void refresh_display()
{
  /*if (mute)
  {
    lcd.noCursor();
    lcd.noBlink();
    lcd.clear();
    lcd.setCursor(0,0);
    lcd.print("    F5X Muses Pre   ");
    lcd.setCursor(6,1);
    lcd.print("Pre Muted");
    lcd.setCursor(0,3);
    if (status_index==0) lcd.print("      Standby       ");
    if (status_index==1) lcd.print("     Pre Amp ON     ");
    if (status_index==2) lcd.print("    Amlifier ON     ");
   
  }
  else
  {
  */
  switch (menu_index){
  case 0:
  {
  lcd.noCursor();
  lcd.noBlink(); 
  lcd.setCursor(0,0);
  lcd.print("    F5X Muses Pre   ");
  lcd.setCursor(0,1);
  lcd.print("  Input" );
  lcd.print(input_index+1);
  lcd.print(": ");
  //lcd.setCursor(11,1);
  lcd.print( input_name[input_index] );
  lcd.setCursor(0,2);
  if (mute){
    
    lcd.print("      Pre Muted     ");
    
    }
  else
  {
  lcd.print("  Volume: " );  
  lcd.print((parameters[0] * STEP) + GAIN);
  lcd.print("   ");
  }
  lcd.setCursor(0,3);
  if (status_index==0) lcd.print("      Standby       ");
  if (status_index==1) lcd.print("     Pre Amp ON     ");
  if (status_index==2) lcd.print("    Amlifier ON     ");

  }
  break;
  case 1:
  case 2:
  case 3:
  case 4:
  {
  
  lcd.setCursor(0,0);
  lcd.cursor();
  lcd.blink();
  lcd.print(" IN");
  lcd.print(menu_index);
  lcd.setCursor(5,0);
  lcd.print( "R-Offset:");
  lcd.print(abs(parameters[(menu_index*10)] * STEP  ));
  if (abs(parameters[(menu_index*10)] * STEP )  < 10.0) lcd.print("  ");
  if (abs(parameters[(menu_index*10)] * STEP ) < 100.0) lcd.print(" ");
  
  lcd.setCursor(0,1);
  lcd.print(" IN");
  lcd.print(menu_index);
  lcd.setCursor(5,1);
  lcd.print( "R+Offset:");
  lcd.print(abs(parameters[(menu_index*10)+1] * STEP ));
  if (abs(parameters[(menu_index*10)] * STEP )  < 10.0) lcd.print("  ");
  if (abs(parameters[(menu_index*10)] * STEP ) < 100.0) lcd.print(" ");

  lcd.setCursor(0,2);
  lcd.print(" IN");
  lcd.print(menu_index);
  lcd.setCursor(5,2);
  lcd.print( "L-Offset:");
  lcd.print(abs(parameters[(menu_index*10)+2] * STEP ));
  if (abs(parameters[(menu_index*10)] * STEP )  < 10.0) lcd.print("  ");
  if (abs(parameters[(menu_index*10)] * STEP ) < 100.0) lcd.print(" ");

  lcd.setCursor(0,3);
  lcd.print(" IN");
  lcd.print(menu_index);
  lcd.setCursor(5,3);
  lcd.print( "L+Offset:");
  lcd.print(abs(parameters[(menu_index*10)+3] * STEP ));
  if (abs(parameters[(menu_index*10)] * STEP )  < 10.0) lcd.print("  ");
  if (abs(parameters[(menu_index*10)] * STEP ) < 100.0) lcd.print(" ");

  //Set the curson on the active field
  lcd.setCursor(0,parameter_index-menu_index*10);

  }
  break; 
   }//end switch
 // }

}//end refresh_display()

//Control Related functions
void select_Input(int val)
{
  muted();
  
  for (int i=0; i <= MAX_INPUT;i++)
  {
  digitalWrite(inputs[i], LOW);
  }
  
  digitalWrite(inputs[val], HIGH);

  
  Update_Volume();
  
  refresh_display();
  
}

void status_Update()

{
  if (status_index==0){
   muted ();
   digitalWrite(PRE_ON, LOW);
   digitalWrite(AMP_ON, LOW);
   lcd.noBacklight();   
  }
  if (status_index==1) {
   muted ();
   lcd.backlight();
   digitalWrite(PRE_ON, HIGH);
   delay(1000);
   digitalWrite(AMP_ON, LOW);
   SetupMuses();
   select_Input(input_index);
   Update_Volume();   
  }
  if (status_index==2) {
   muted ();
   digitalWrite(PRE_ON, HIGH);
   digitalWrite(AMP_ON, HIGH);
   delay(1000);
   Update_Volume(); 
  }
  
}

#ifdef DEBUG
void printpar() //Debug function
{
  for (int i=0; i< MAX_PARAMETER+1;i++)
    {
      Serial.print(parameters[i]);
      Serial.print("\t");
    }
    
  Serial.print("\n");
}
#endif
void setup()
 
 {
  // initialize the LCD
    lcd.init();
    lcd.backlight();
    MusesR.begin();
    MusesL.begin();
    lcd.setContrast(100);
     #ifdef DEBUG
      Serial.begin(9600);
     #endif
    main_knob= new ClickEncoder(A0, A1, A2,2);
    Timer1.initialize(100);
    Timer1.attachInterrupt(timerIsr); 
    main_knob->setAccelerationEnabled(true);
    
    pinMode(INPUT_1, OUTPUT);
    pinMode(INPUT_2, OUTPUT);
    pinMode(INPUT_3, OUTPUT);
    pinMode(INPUT_4, OUTPUT);
    pinMode(PRE_ON, OUTPUT);
    pinMode(AMP_ON, OUTPUT);

    IrReceiver.begin(IR_RECEIVE_PIN, ENABLE_LED_FEEDBACK);
    
    lcd.setCursor(0,0);
    lcd.print("    F5X Muses Pre   ");
    lcd.setCursor(0,1);
    lcd.print("     v1.0 Muses     ");
    lcd.setCursor(0,2);
    lcd.print(date);
    lcd.setCursor(0,3);
    lcd.print(build); 
    delay(4000);
    lcd.clear();
    readParameters();
    initialize_parameters();
    status_Update();
    refresh_display(); 
}


void loop() {

  main_button();
  remote();
  
  if (status_index !=0) 
  
  {main_knobf();
   
  }
 }
